local Recount2 = Recount2

local setfenv = setfenv
local loadstring = loadstring
local select = select

local combatants

function Recount2:InitExpressions()
	combatants = Recount2.Data.Combatants
end

--:CompileExpression
--Compiles a mathematical expression into a state machine stack
local fullTable = {}
function Recount2:CompileExpression(str)
	--return codeStack
	code, msg = loadstring("return "..str)
	setfenv(code, fullTable)
	return code
end


function Recount2:AddExpressionFunction(funcName, func)
	fullTable[funcName] = func
end

function Recount2:AddExpressionConstant(constantName, value)
	fullTable[constantName] = value
end

function Recount2:AddExpressionConstants(...)
	local count = select("#", ...)
	for i = 1, count, 2 do
		local constantName, value = select(i,...)
		fullTable[constantName] = value
	end
end

local function Add(a, b)
	return a + b
end

local function Sub(a, b)
	return a - b
end

local function Mul(a, b)
	return a * b
end

local function Div(a, b)
	return a / b
end

local function Concat(a, b)
	return a..b
end

local function LessThan(a, b)
	return a<b
end

local function GreaterThan(a, b)
	return a>b
end

local function LessThanOrEqual(a, b)
	return a<=b
end

local function GreaterThanOrEqual(a, b)
	return a>=b
end

local function Equal(a, b)
	return a==b
end

local function NotEqual(a, b)
	return a~=b
end

local function AndOp(a, b)
	return a and b
end

local function OrOp(a, b)
	return a or b
end

local function IfFunc(a, b, c)
	if a == '$nil^' or not a then
		return c
	end
	return b
end

local function Negate(a)
	return -a
end

local string_format = string.format
local function Short(num)
	if num>1000000 then
		return string_format("%.2fm", num/1000000)
	elseif num>1000 then
		return string_format("%.1fk", num/1000)
	end
	return tostring(num)
end

function Recount2:ClearExpressionNamespace()
	for k in pairs(fullTable) do
		fullTable[k] = nil
	end


	Recount2:AddExpressionFunction("Negate", Negate)
	
	Recount2:AddExpressionFunction("Add", Add)	
	Recount2:AddExpressionFunction("Sub", Sub)	
	Recount2:AddExpressionFunction("Mul", Mul)
	Recount2:AddExpressionFunction("Div", Div)	
	Recount2:AddExpressionFunction("Mod", math.fmod)
	

	Recount2:AddExpressionFunction("Max", math.max)
	Recount2:AddExpressionFunction("Min", math.min)

	Recount2:AddExpressionFunction("Log", math.log)
	Recount2:AddExpressionFunction("Log10", math.log10)
	Recount2:AddExpressionFunction("Exp", math.exp)
	Recount2:AddExpressionFunction("Sqrt", math.sqrt)
	Recount2:AddExpressionFunction("Power", math.pow)

	Recount2:AddExpressionFunction("Abs", math.abs)
	Recount2:AddExpressionFunction("Ceil", math.ceil)
	Recount2:AddExpressionFunction("Floor", math.floor)

	Recount2:AddExpressionFunction("Rad", math.rad)
	Recount2:AddExpressionFunction("Deg", math.deg)
	
	Recount2:AddExpressionFunction("Sin", math.sin)
	Recount2:AddExpressionFunction("Cos", math.cos)
	Recount2:AddExpressionFunction("Tan", math.tan)
	Recount2:AddExpressionFunction("ArcSin", math.asin)
	Recount2:AddExpressionFunction("ArcCos", math.acos)
	Recount2:AddExpressionFunction("ArcTan", math.atan)

	Recount2:AddExpressionFunction("Short", Short)

	--String operations
	Recount2:AddExpressionFunction("Upper", string.upper)
	Recount2:AddExpressionFunction("Lower", string.lower)
	Recount2:AddExpressionFunction("Reverse", string.reverse)
	Recount2:AddExpressionFunction("Len", string.len)
	Recount2:AddExpressionFunction("SubString", string.sub)
	Recount2:AddExpressionFunction("Concat", Concat)

	--Boolean	
	Recount2:AddExpressionFunction("And", AndOp)	
	Recount2:AddExpressionFunction("Or", OrOp)

	Recount2:AddExpressionFunction("If", IfFunc)

	Recount2:AddExpressionConstant("nil", nil)
	Recount2:AddExpressionConstant("true", true)
	Recount2:AddExpressionConstant("false", false)
	Recount2:AddExpressionConstant("pi", math.pi)
	Recount2:AddExpressionConstant("huge", math.huge)
end

Recount2:ClearExpressionNamespace()


function Recount2:ProcessExpression(expression)
	local cExpression = Recount2:CompileExpression(expression)
	return cExpression()
end

function Recount2:SetGUIDInfo(guid)
	local attrTable = combatants[guid].attrTable

	fullTable["cName"]	= attrTable.name
	fullTable["cClass"]	= attrTable.class
	fullTable["cRace"]	= attrTable.race
	fullTable["cClassification"]	= attrTable.classification
	fullTable["cType"]	= attrTable.creatureType
	fullTable["cFamily"]	= attrTable.creatureFamily
	fullTable["cFaction"]	= attrTable.faction
	fullTable["cSex"]	= attrTable.sex
	fullTable["cLevel"]	= attrTable.level

	return attrTable.name
end

function Recount2.SetCombatantConstants(entry)
	fullTable["cCount"] = entry.count or 0
	fullTable["cTotal"] = entry.total or 0
	fullTable["cMin"]	= entry.min or 0
	fullTable["cMax"]	= entry.max or 0
	fullTable["cResist"]= entry.resist or 0
	fullTable["cAbsorb"]= entry.absorb or 0
	fullTable["cBlock"]	= entry.block or 0
	fullTable["cOver"]	= entry.over or 0
end