Recount2 = LibStub("AceAddon-3.0"):NewAddon("Recount 2", "AceConsole-3.0" )
local Recount2 = Recount2

--Are we in debug mode?
Recount2.Debug = true

--Setup subnamespace for widgets
Recount2.Widgets = {}

--Setup subnamespace for translation tables
Recount2.TranslationTables = {}

local GetTime = GetTime
local tinsert = table.insert
local tremove = table.remove

local function EventHandler(self, event, ...)
	local _ = Recount2.EventHandlers[event] ~= nil and Recount2.EventHandlers[event](Recount2, event, ...)
end

local ScheduledID
local ScheduledFuncs
local ScheduledFor
local ScheduledRepeat
local ScheduledArgs

local function InitData()
	for _, v in pairs(Recount2.mainListWindows) do
		v:ResetCurFilter()
	end

	Recount2:InitBuffer()
	Recount2:InitDataRecorder()
	Recount2:InitGUIDTracking()
	Recount2:InitTracker()
	Recount2:InitExpressions()

	Recount2:AddActiveDataset("Overall", Recount2.Data.Overall)
	Recount2:AddActiveDataset("Current", Recount2.Data.Fight[0])
end

function Recount2:ResetData(first)
	data = self:GetTable()
	Recount2PerCharDB.combatData = data
	self.Data = data
	data.Combatants = self:GetTable()
	data.NPCData = self:GetTable()
	data.Overall = self:GetTable()
	data.Fight = self:GetTable()
	data.Fight[0] = self:GetTable()

	InitData()

	if not first then
		Recount2:Print("Data Reset")
	end
end

function Recount2:OnInitialize()
	self.filter = {}
	self.viewTable = {}
	self.viewCompress = {}
	self.activeDatasets = {}

	self.mainListWindows = {}
	self.detailWindows = {}
	self.configListWindows = {}
	
	self.EventFrame = CreateFrame("FRAME", nil, nil)
	self.EventFrame:SetScript("OnEvent", EventHandler)
	self.EventHandlers = {}
	self.EventHandlers["COMBAT_LOG_EVENT_UNFILTERED"] = self.CombatLogEvent

	self.ScheduledFuncs = {}
	self.ScheduledArgs = {}
	self.ScheduledFor = {}
	self.ScheduledRepeat = {}
	self.ScheduledID = {}

	ScheduledFuncs = self.ScheduledFuncs
	ScheduledFor = self.ScheduledFor
	ScheduledRepeat = self.ScheduledRepeat
	ScheduledArgs = self.ScheduledArgs
	ScheduledID = self.ScheduledID


	self.options = {}
	self.options.performanceMode = false

	self.dataVersion = 1

	--Init Data if there is none
	if type(Recount2PerCharDB)=="table" and Recount2PerCharDB.version==self.dataVersion then
		self.Data = Recount2PerCharDB.combatData

		InitData()
	else
		Recount2PerCharDB = {}
		Recount2PerCharDB.version = self.dataVersion
		Recount2PerCharDB.filterEvents = {
			['Damage'] = true,
			['Healing'] = true,
			['Misses'] = true,
			['Power Gains'] = true,
		}
		Recount2PerCharDB.filterAffiliation = {
			['Mine'] = true,
			['Party'] = true, 
			['Raid'] = true,
			['Others'] = true,
		}
		self:ResetData(true)
	end

	self.filterEvents = Recount2PerCharDB.filterEvents
	self.filterAffiliation = Recount2PerCharDB.filterAffiliation

	if type(Recount2DB)~="table" then
		Recount2DB = {}
		Recount2DB.dataVersion = self.dataVersion

		Recount2DB.Filters = Recount2.DefaultFilters
	else
		--Check if any filters from the default appear to be missing
		Recount2:CheckFilters()
	end
	self.SharedData = Recount2DB
	

	Recount2:RegisterChatCommand("recount2", Recount2.HandleConsole)
end

function Recount2:RegisterEvent(event, func)
	self.EventHandlers[event] = func
	self.EventFrame:RegisterEvent(event)
end

function Recount2:UnregisterEvent(event)
	self.EventHandlers[event] = nil
	self.EventFrame:UnregisterEvent(event)
end

function Recount2:AddMainWindow()
	for k, v in ipairs(Recount2.mainListWindows) do
		if not v:IsShown() then
			Recount2:UpdateMainListWindow(k, Recount2.SharedData.Filters[1], Recount2.SharedData.Filters)
			Recount2:Print("Showing Window "..k)
			return
		end
	end
	Recount2:UpdateMainListWindow(1+#Recount2.mainListWindows, Recount2.SharedData.Filters[1], Recount2.SharedData.Filters)
	Recount2:Print("Showing Window "..#Recount2.mainListWindows)
end

function Recount2:HideMainWindows()
	for k, v in pairs(Recount2.mainListWindows) do
		if v:IsShown() then
			v:Hide()
			return
		end
	end
end


function Recount2.HandleConsole(args)
	local arg, nextPos = Recount2:GetArgs(args)

	if arg then
		arg = arg:lower()
	end

	if arg=="show" then
		local windowName = Recount2:GetArgs(args, 1, nextPos)
		if windowName then
			if Recount2.mainListWindows[windowName]==nil or not Recount2.mainListWindows[windowName]:IsShown() then
				Recount2:Print("Showing Window "..windowName)
				Recount2:UpdateMainListWindow(windowName, Recount2.SharedData.Filters[1], Recount2.SharedData.Filters)
				return
			end
		else
			Recount2:AddMainWindow()
		end
	elseif arg=="hide" then
		local windowName = Recount2:GetArgs(args, 1, nextPos)
		if windowName then
			if Recount2.mainListWindows[windowName] and Recount2.mainListWindows[windowName]:IsShown() then
				Recount2.mainListWindows[windowName]:Hide()
				return
			end
		else
			Recount2:HideMainWindows()
		end
	elseif arg=="reset" then
		Recount2:ResetData()
	elseif arg=="performance" then
		Recount2.options.performanceMode = not Recount2.options.performanceMode
		if Recount2.inCombat then
			if Recount2.options.performanceMode then
				Recount2:StopViewUpdating()
			elseif not Recount2.options.performanceMode then
				Recount2:StartViewUpdating()
			end
		end
		
		if Recount2.options.performanceMode then
			Recount2:Print("Performance Mode Enabled")
		else
			Recount2:Print("Performance Mode Disabled")
		end
	else
		Recount2:Print("Console Commands")
		Recount2:Print("hide [windowName] - Hides all list windows or the explicitly specified list window name")
		Recount2:Print("reset - Resets all data in Recount 2")
		Recount2:Print("show [windowName] - Show a new list window or the explicitly specified list window name")
		if Recount2.options.performanceMode then
			Recount2:Print("performance - Disables performance mode renabling updates during combat (currently enabled)")
		else
			Recount2:Print("performance - Enables performance mode disabling updates during combat (currently disabled)")
		end
	end
end

function Recount2:OnEnable()
	self.EventFrame:SetScript("OnEvent", EventHandler)

	self:UpdateMainListWindow(1, self.SharedData.Filters[1], self.SharedData.Filters)
	
	self.EventFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterGUIDTracking()

	if self.initialMemory == nil then
		collectgarbage("collect")
		UpdateAddOnMemoryUsage()
		self.initialMemory = GetAddOnMemoryUsage("Recount2")
	end
end

function Recount2:OnDisable()
	self.EventFrame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:UnregisterGUIDTracking()
end

function Recount2:PrintStats()
	self:Print("Initial Memory:", self.initialMemory)
	local nCombatants = 0
	for k,v in pairs(self.Data.Combatants) do
		nCombatants = nCombatants + 1
	end
	UpdateAddOnMemoryUsage()
	local updatedMemory = GetAddOnMemoryUsage("Recount2") - self.initialMemory
	self:Print("Additional Memory:", updatedMemory)
	self:Print("Number of Combatants:", nCombatants)
	self:Print("Memory per Combatant:", updatedMemory/nCombatants)
end

function Recount2:ScheduleTimer(id, func, time, ...)
	tinsert(ScheduledID, id)
	tinsert(ScheduledFuncs, func)
	tinsert(ScheduledFor, GetTime()+time)
	tinsert(ScheduledRepeat, -1)

	if self.lowestTime==nil or self.lowestTime>ScheduledFor[#ScheduledFor] then
		self.lowestTime=ScheduledFor[#ScheduledFor]
	end

	local args = self:GetTable()
	local nArgs = select("#", ...)

	for i = 1, nArgs do
		local curArg = select(i, ...)
		tinsert(args, curArg)
	end
	tinsert(ScheduledArgs, args)

	self.EventFrame:SetScript("OnUpdate", self.HandleTimers)
end

function Recount2:ScheduleRepeatingTimer(id, func, time, ...)
	tinsert(ScheduledID, id)
	tinsert(ScheduledFuncs, func)
	tinsert(ScheduledFor, GetTime()+time)
	tinsert(ScheduledRepeat, time)

	if self.lowestTime==nil or self.lowestTime>ScheduledFor[#ScheduledFor] then
		self.lowestTime=ScheduledFor[#ScheduledFor]
	end

	local args = self:GetTable()
	local nArgs = select("#", ...)

	for i = 1, nArgs do
		local curArg = select(i, ...)
		tinsert(args, curArg)
	end
	tinsert(ScheduledArgs, args)
	self.EventFrame:SetScript("OnUpdate", self.HandleTimers)
end

function Recount2:RemoveTimer(id)
	for k, v in ipairs(ScheduledID) do
		if v==id then
			tremove(ScheduledID, k)
			tremove(ScheduledFor, k)
			tremove(ScheduledFuncs, k)
			tremove(ScheduledRepeat, k)
			Recount2:FreeTable(ScheduledArgs[k])
			tremove(ScheduledArgs, k)
			return
		end
	end
end

local removeLater = {}
function Recount2:HandleTimers()
	local curTime = GetTime()

	if Recount2.dataInBuffer then
		Recount2:ProcessBuffer()

		if Recount2.dataInBuffer==false and #ScheduledFor==0 then
			Recount2.EventFrame:SetScript("OnUpdate", nil)
			return
		end
	elseif #ScheduledFor==0 then
		Recount2.EventFrame:SetScript("OnUpdate", nil)
		return
	end

	if Recount2.lowestTime==nil or Recount2.lowestTime>curTime then
		return
	end

	local lowestTime = nil
	

	for k, v in ipairs(ScheduledFor) do
		if v<curTime then
			ScheduledFuncs[k](unpack(ScheduledArgs[k]))

			if ScheduledRepeat[k]>=0 then
				ScheduledFor[k] = curTime + ScheduledRepeat[k]
				if lowestTime == nil or lowestTime>ScheduledFor[k] then
					lowestTime = ScheduledFor[k]
				end
			else				
				tinsert(removeLater, 1, k)
			end
		elseif lowestTime == nil or lowestTime>v then
			lowestTime = v
		end
	end

	for k, v in ipairs(removeLater) do
		tremove(ScheduledID, v)
		tremove(ScheduledFor, v)
		tremove(ScheduledFuncs, v)
		tremove(ScheduledRepeat, v)
		Recount2:FreeTable(ScheduledArgs[v])
		tremove(ScheduledArgs, v)
		removeLater[k] = nil
	end

	Recount2.lowestTime = lowestTime

	if Recount2.dataInBuffer==false and #ScheduledFor==0 then
		Recount2.EventFrame:SetScript("OnUpdate", nil)
	end
end